@@ -1,10 +1,11 @@
 from __future__ import annotations
 
 from copy import deepcopy
-from typing import Any, Iterable, Sequence
+from typing import Any, TypeVar, TYPE_CHECKING
+from collections.abc import Iterable, Sequence
 
 from django import forms
-from django.db.models import Field
+from django.db.models import Field, Model
 from django.contrib import admin
 from django.contrib.admin.options import BaseModelAdmin, InlineModelAdmin, flatten_fieldsets
 from django.contrib.contenttypes.admin import GenericStackedInline, GenericTabularInline
@@ -25,8 +26,15 @@ from modeltranslation.utils import (
 from modeltranslation.widgets import ClearableWidgetWrapper
 from modeltranslation._typing import _ListOrTuple
 
+if TYPE_CHECKING:
+    # We depend here or `django-stubs` internal `_FieldsetSpec`,
+    # in case it changes, change import or define this internally.
+    from django.contrib.admin.options import _FieldsetSpec
 
-class TranslationBaseModelAdmin(BaseModelAdmin):
+_ModelT = TypeVar("_ModelT", bound=Model)
+
+
+class TranslationBaseModelAdmin(BaseModelAdmin[_ModelT]):
     _orig_was_required: dict[str, bool] = {}
     both_empty_values_fields = ()
 
@@ -36,11 +44,11 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         self._patch_prepopulated_fields()
 
     def _get_declared_fieldsets(
-        self, request: HttpRequest, obj: Any | None = None
-    ) -> _ListOrTuple[tuple[str | None, dict[str, Any]]] | None:
+        self, request: HttpRequest, obj: _ModelT | None = None
+    ) -> _FieldsetSpec | None:
         # Take custom modelform fields option into account
         if not self.fields and hasattr(self.form, "_meta") and self.form._meta.fields:
-            self.fields = self.form._meta.fields
+            self.fields = self.form._meta.fields  # type: ignore[assignment]
         # takes into account non-standard add_fieldsets attribute used by UserAdmin
         fieldsets = (
             self.add_fieldsets
@@ -53,17 +61,24 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
             return [(None, {"fields": self.replace_orig_field(self.get_fields(request, obj))})]
         return None
 
+    def _patch_fieldsets(self, fieldsets: _FieldsetSpec) -> _FieldsetSpec:
+        fieldsets_new = list(fieldsets)
+        for name, dct in fieldsets:
+            if "fields" in dct:
+                dct["fields"] = self.replace_orig_field(dct["fields"])
+        return fieldsets_new
+
     def formfield_for_dbfield(
         self, db_field: Field, request: HttpRequest, **kwargs: Any
-    ) -> forms.Field:
-        field = super().formfield_for_dbfield(db_field, request, **kwargs)
-        self.patch_translation_field(db_field, field, request, **kwargs)  # type: ignore[arg-type]
-        return field  # type: ignore[return-value]
+    ) -> forms.Field | None:
+        if field := super().formfield_for_dbfield(db_field, request, **kwargs):
+            self.patch_translation_field(db_field, field, request, **kwargs)
+        return field
 
     def patch_translation_field(
         self, db_field: Field, field: forms.Field, request: HttpRequest, **kwargs: Any
     ) -> None:
-        if db_field.name in self.trans_opts.fields:
+        if db_field.name in self.trans_opts.all_fields:
             if field.required:
                 field.required = False
                 field.blank = True
@@ -77,6 +92,8 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
             pass
         else:
             orig_formfield = self.formfield_for_dbfield(orig_field, request, **kwargs)
+            if orig_formfield is None:
+                return
             field.widget = deepcopy(orig_formfield.widget)
             attrs = field.widget.attrs
             # if any widget attrs are defined on the form they should be copied
@@ -140,8 +157,8 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
             exclude = tuple()
         if exclude:
             exclude_new = tuple(exclude)
-            return exclude_new + tuple(self.trans_opts.fields.keys())
-        return tuple(self.trans_opts.fields.keys())
+            return exclude_new + tuple(self.trans_opts.all_fields.keys())
+        return tuple(self.trans_opts.all_fields.keys())
 
     def replace_orig_field(self, option: Iterable[str | Sequence[str]]) -> _ListOrTuple[str]:
         """
@@ -171,34 +188,23 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         if option:
             option_new = list(option)
             for opt in option:
-                if opt in self.trans_opts.fields:
+                if opt in self.trans_opts.all_fields:
                     index = option_new.index(opt)
                     option_new[index : index + 1] = get_translation_fields(opt)  # type: ignore[arg-type]
                 elif isinstance(opt, (tuple, list)) and (
-                    [o for o in opt if o in self.trans_opts.fields]
+                    [o for o in opt if o in self.trans_opts.all_fields]
                 ):
                     index = option_new.index(opt)
                     option_new[index : index + 1] = self.replace_orig_field(opt)
             option = option_new
         return option  # type: ignore[return-value]
 
-    def _patch_fieldsets(
-        self, fieldsets: _ListOrTuple[tuple[str | None, dict[str, Any]]]
-    ) -> _ListOrTuple[tuple[str | None, dict[str, Any]]]:
-        if fieldsets:
-            fieldsets_new = list(fieldsets)
-            for name, dct in fieldsets:
-                if "fields" in dct:
-                    dct["fields"] = self.replace_orig_field(dct["fields"])
-            fieldsets = fieldsets_new
-        return fieldsets
-
     def _patch_prepopulated_fields(self) -> None:
         def localize(sources: Sequence[str], lang: str) -> tuple[str, ...]:
             "Append lang suffix (if applicable) to field list"
 
             def append_lang(source: str) -> str:
-                if source in self.trans_opts.fields:
+                if source in self.trans_opts.all_fields:
                     return build_localized_fieldname(source, lang)
                 return source
 
@@ -206,7 +212,7 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
 
         prepopulated_fields: dict[str, Sequence[str]] = {}
         for dest, sources in self.prepopulated_fields.items():
-            if dest in self.trans_opts.fields:
+            if dest in self.trans_opts.all_fields:
                 for lang in mt_settings.AVAILABLE_LANGUAGES:
                     key = build_localized_fieldname(dest, lang)
                     prepopulated_fields[key] = localize(sources, lang)
@@ -216,17 +222,17 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         self.prepopulated_fields = prepopulated_fields
 
     def _get_form_or_formset(
-        self, request: HttpRequest, obj: Any | None, **kwargs: Any
+        self, request: HttpRequest, obj: Model | None, **kwargs: Any
     ) -> dict[str, Any]:
         """
         Generic code shared by get_form and get_formset.
         """
-        exclude = self.get_exclude(request, obj)
+        exclude = self.get_exclude(request, obj)  # type: ignore[arg-type]
         if exclude is None:
             exclude = []
         else:
             exclude = list(exclude)
-        exclude.extend(self.get_readonly_fields(request, obj))
+        exclude.extend(self.get_readonly_fields(request, obj))  # type: ignore[arg-type]
         if not exclude and hasattr(self.form, "_meta") and self.form._meta.exclude:
             # Take the custom ModelForm's Meta.exclude into account only if the
             # ModelAdmin doesn't define its own.
@@ -240,8 +246,8 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         return kwargs
 
     def _get_fieldsets_pre_form_or_formset(
-        self, request: HttpRequest, obj: Any | None = None
-    ) -> _ListOrTuple[tuple[str | None, dict[str, Any]]] | None:
+        self, request: HttpRequest, obj: _ModelT | None = None
+    ) -> _FieldsetSpec | None:
         """
         Generic get_fieldsets code, shared by
         TranslationAdmin and TranslationInlineModelAdmin.
@@ -249,7 +255,7 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         return self._get_declared_fieldsets(request, obj)
 
     def _get_fieldsets_post_form_or_formset(
-        self, request: HttpRequest, form: type[forms.ModelForm], obj: Any | None = None
+        self, request: HttpRequest, form: type[forms.ModelForm], obj: _ModelT | None = None
     ) -> list:
         """
         Generic get_fieldsets code, shared by
@@ -273,14 +279,14 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         if exclude_languages:
             excl_languages = exclude_languages
         exclude = []
-        for orig_fieldname, translation_fields in self.trans_opts.fields.items():
+        for orig_fieldname, translation_fields in self.trans_opts.all_fields.items():
             for tfield in translation_fields:
                 if tfield.language in excl_languages and tfield not in exclude:
                     exclude.append(tfield)
         return tuple(exclude)
 
     def get_readonly_fields(
-        self, request: HttpRequest, obj: Any | None = None
+        self, request: HttpRequest, obj: _ModelT | None = None
     ) -> _ListOrTuple[str]:
         """
         Hook to specify custom readonly fields.
@@ -288,7 +294,7 @@ class TranslationBaseModelAdmin(BaseModelAdmin):
         return self.replace_orig_field(self.readonly_fields)
 
 
-class TranslationAdmin(TranslationBaseModelAdmin, admin.ModelAdmin):
+class TranslationAdmin(TranslationBaseModelAdmin[_ModelT], admin.ModelAdmin[_ModelT]):
     # TODO: Consider addition of a setting which allows to override the fallback to True
     group_fieldsets = False
 
@@ -301,7 +307,7 @@ class TranslationAdmin(TranslationBaseModelAdmin, admin.ModelAdmin):
             editable_new = list(self.list_editable)
             display_new = list(self.list_display)
             for field in self.list_editable:
-                if field in self.trans_opts.fields:
+                if field in self.trans_opts.all_fields:
                     index = editable_new.index(field)
                     display_index = display_new.index(field)
                     translation_fields = get_translation_fields(field)
@@ -350,13 +356,13 @@ class TranslationAdmin(TranslationBaseModelAdmin, admin.ModelAdmin):
             )
 
             temp_fieldsets = {}
-            for orig_field, trans_fields in self.trans_opts.fields.items():
+            for orig_field, trans_fields in self.trans_opts.all_fields.items():
                 trans_fieldnames = [f.name for f in sorted(trans_fields, key=lambda x: x.name)]
                 if any(f in trans_fieldnames for f in flattened_fieldsets):
                     # Extract the original field's verbose_name for use as this
                     # fieldset's label - using gettext_lazy in your model
                     # declaration can make that translatable.
-                    label = self.model._meta.get_field(orig_field).verbose_name.capitalize()
+                    label = self.model._meta.get_field(orig_field).verbose_name.capitalize()  # type: ignore[union-attr]
                     temp_fieldsets[orig_field] = (
                         label,
                         {"fields": trans_fieldnames, "classes": ("mt-fieldset",)},
@@ -374,14 +380,12 @@ class TranslationAdmin(TranslationBaseModelAdmin, admin.ModelAdmin):
         return fieldsets
 
     def get_form(
-        self, request: HttpRequest, obj: Any | None = None, **kwargs: Any
+        self, request: HttpRequest, obj: _ModelT | None = None, **kwargs: Any
     ) -> type[forms.ModelForm]:
         kwargs = self._get_form_or_formset(request, obj, **kwargs)
         return super().get_form(request, obj, **kwargs)
 
-    def get_fieldsets(
-        self, request: HttpRequest, obj: Any | None = None
-    ) -> _ListOrTuple[tuple[str | None, dict[str, Any]]]:
+    def get_fieldsets(self, request: HttpRequest, obj: _ModelT | None = None) -> _FieldsetSpec:
         return self._get_fieldsets_pre_form_or_formset(request, obj) or self._group_fieldsets(
             self._get_fieldsets_post_form_or_formset(
                 request, self.get_form(request, obj, fields=None), obj
@@ -389,41 +393,57 @@ class TranslationAdmin(TranslationBaseModelAdmin, admin.ModelAdmin):
         )
 
 
-class TranslationInlineModelAdmin(TranslationBaseModelAdmin, InlineModelAdmin):
+_ChildModelT = TypeVar("_ChildModelT", bound=Model)
+_ParentModelT = TypeVar("_ParentModelT", bound=Model)
+
+
+class TranslationInlineModelAdmin(
+    TranslationBaseModelAdmin[_ChildModelT], InlineModelAdmin[_ChildModelT, _ParentModelT]
+):
     def get_formset(
-        self, request: HttpRequest, obj: Any | None = None, **kwargs: Any
+        self, request: HttpRequest, obj: _ParentModelT | None = None, **kwargs: Any
     ) -> type[BaseInlineFormSet]:
         kwargs = self._get_form_or_formset(request, obj, **kwargs)
         return super().get_formset(request, obj, **kwargs)
 
-    def get_fieldsets(self, request: HttpRequest, obj: Any | None = None):
+    def get_fieldsets(self, request: HttpRequest, obj: _ChildModelT | None = None):
         # FIXME: If fieldsets are declared on an inline some kind of ghost
         # fieldset line with just the original model verbose_name of the model
         # is displayed above the new fieldsets.
         declared_fieldsets = self._get_fieldsets_pre_form_or_formset(request, obj)
         if declared_fieldsets:
             return declared_fieldsets
-        form = self.get_formset(request, obj, fields=None).form
+        form = self.get_formset(request, obj, fields=None).form  # type: ignore[arg-type]
         return self._get_fieldsets_post_form_or_formset(request, form, obj)
 
 
-class TranslationTabularInline(TranslationInlineModelAdmin, admin.TabularInline):
+class TranslationTabularInline(
+    TranslationInlineModelAdmin[_ChildModelT, _ParentModelT],
+    admin.TabularInline[_ChildModelT, _ParentModelT],
+):
     pass
 
 
-class TranslationStackedInline(TranslationInlineModelAdmin, admin.StackedInline):
+class TranslationStackedInline(
+    TranslationInlineModelAdmin[_ChildModelT, _ParentModelT],
+    admin.StackedInline[_ChildModelT, _ParentModelT],
+):
     pass
 
 
-class TranslationGenericTabularInline(TranslationInlineModelAdmin, GenericTabularInline):
+class TranslationGenericTabularInline(
+    TranslationInlineModelAdmin[_ChildModelT, _ParentModelT], GenericTabularInline
+):
     pass
 
 
-class TranslationGenericStackedInline(TranslationInlineModelAdmin, GenericStackedInline):
+class TranslationGenericStackedInline(
+    TranslationInlineModelAdmin[_ChildModelT, _ParentModelT], GenericStackedInline
+):
     pass
 
 
-class TabbedDjangoJqueryTranslationAdmin(TranslationAdmin):
+class TabbedDjangoJqueryTranslationAdmin(TranslationAdmin[_ModelT]):
     """
     Convenience class which includes the necessary media files for tabbed
     translation fields. Reuses Django's internal jquery version.
@@ -441,7 +461,7 @@ class TabbedDjangoJqueryTranslationAdmin(TranslationAdmin):
         }
 
 
-class TabbedExternalJqueryTranslationAdmin(TranslationAdmin):
+class TabbedExternalJqueryTranslationAdmin(TranslationAdmin[_ModelT]):
     """
     Convenience class which includes the necessary media files for tabbed
     translation fields. Loads recent jquery version from a cdn.
